#Standard python libraries
import numpy as np
import os
import yaml
import warnings
import numbers
import itertools
from datetime import datetime
import time
import matplotlib.pyplot as plt
from scipy.sparse import csr_matrix

from .eigenstates import PolymerVibrations
from .base_class import DataOrganizer

class EigenGenerator(PolymerVibrations,DataOrganizer):
    """This class generates the eigenvalues and eigenvectors
        needed for the specified system in the params file.  This
        class should be initialized once at the beginning of any calculation 
        run, and the generate and save methods should both be called once."""
    def __init__(self,parameter_file_path,*,check_convergence=True,
                 qr_flag=False,mask_by_occupation_num=True):
        parameter_file = os.path.join(parameter_file_path,'params.yaml')
        super().__init__(parameter_file,qr_flag=qr_flag,
                         mask_by_occupation_num=mask_by_occupation_num)
        self.base_path = parameter_file_path
        self.load_params()

        self.set_diagrams_and_manifolds()

        self.check_convergence = check_convergence
            
        try:
            self.load_eigenvalues_and_vectors()
        except (FileNotFoundError, KeyError):
            self.generate_eigenvalues_and_vectors_with_tolerance()
            self.save_eigenvalues_and_vectors()

    def set_diagrams_and_manifolds(self):
        if len(self.energies) == 2:
            self.diagrams = ['GSB', 'SE']
            self.manifolds = ['GSM','SEM']
            self.diagrams_by_manifold = {'GSM':['GSB'],'SEM':['SE']}
        elif len(self.energies) == 3:
            self.diagrams = ['GSB','SE','ESA']
            self.manifolds = ['GSM','SEM','DEM']
            self.diagrams_by_manifold = {'GSM':['GSB'],'SEM':['SE','ESA']}

    def recalculate_eigenvalues_and_vectors(self):
        self.truncation_size = self.params['initial truncation size']
        self.generate_eigenvalues_and_vectors_with_tolerance()
        self.save_eigenvalues_and_vectors()

    def generate_eigenvalues_and_vectors_with_tolerance(self,*,save_intermediate = False):
        eig_tol = self.params['eigenvalue precision']
        rel_err = 1 if self.check_convergence else 0
        N = self.params['number eigenvalues']
        if N == 'full':
            # If using eigh instead of eigsh
            rel_err = 0
        max_iter = 20
        self.generate_eigenvalues_and_vectors()
        
        old_eigvals = np.hstack((self.eigenvalues[0],self.eigenvalues[1]))
        if 'DEM' in self.manifolds:
            old_eigvals = np.hstack((old_eigvals,self.eigenvalues[2]))
        iter_count = 0
        while rel_err > eig_tol:
            if save_intermediate:
                eigval_save_name = os.path.join(self.base_path,'eigenvalues_eigs_truncation{}.npz'.format(self.truncation_size))
                eigval_dict = {key:arr for key,arr in zip(self.manifolds,self.eigenvalues)}
                np.savez(eigval_save_name,**eigval_dict)
            self.truncation_size += 1
            print(self.truncation_size)
            self.set_vibrations()
            self.generate_eigenvalues_and_vectors()
            new_eigvals = np.hstack((self.eigenvalues[0],self.eigenvalues[1]))
            if 'DEM' in self.manifolds:
                new_eigvals = np.hstack((new_eigvals,self.eigenvalues[2]))
            
            # There will always be eigenvalues with value 0, hence the plus 1
            rel_err = np.max(np.abs((new_eigvals - old_eigvals)/(new_eigvals+1)))
            old_eigvals = new_eigvals
            iter_count += 1
            if iter_count > max_iter:
                warnings.warn('Eigenvalues did not converge to requested tolerance')
                break

    def generate_eigenvalues_and_vectors_with_tolerance_eig(self,*,save_intermediate=False):
        eig_tol = self.params['eigenvalue precision']
        rel_err = 1 if self.check_convergence else 0
        N = self.params['number eigenvalues']
        max_iter = 20
        self.generate_eigenvalues_and_vectors_eig()
        
        old_eigvals = np.hstack((self.eigenvalues[0],self.eigenvalues[1]))
        if 'DEM' in self.manifolds:
            old_eigvals = np.hstack((old_eigvals,self.eigenvalues[2]))
        
        iter_count = 0
        while rel_err > eig_tol:
            if save_intermediate:
                eigval_save_name = os.path.join(self.base_path,'eigenvalues_eig_truncation{}.npz'.format(self.truncation_size))
                eigval_dict = {key:arr for key,arr in zip(self.manifolds,self.eigenvalues)}
                np.savez(eigval_save_name,**eigval_dict)
            self.truncation_size += 2
            self.set_vibrations()
            self.generate_eigenvalues_and_vectors_eig()
            new_eigvals = np.hstack((self.eigenvalues[0],self.eigenvalues[1]))
            
            if 'DEM' in self.manifolds:
                new_eigvals = np.hstack((new_eigvals,self.eigenvalues[2]))
            
            # There will always be eigenvalues with value 0, hence the plus 1
            rel_err = np.max(np.abs((new_eigvals - old_eigvals)/(new_eigvals+1)))
            old_eigvals = new_eigvals
            iter_count += 1
            if iter_count > max_iter:
                warnings.warn('Eigenvalues did not converge to requested tolerance')
                break

    def generate_eigenvalues_and_vectors(self):
        """Use the methods from PolymerVibrations to construct the 
            eigenvalues and eigenmatrices for the GSM, SEM, and if 
            applicable, the DEM
"""
        self.eigenvalues = []
        self.eigenvectors = []
        self.time_to_find_eigensystem = []

        N = self.params['number eigenvalues']

        for i in range(len(self.energies)):
            t0 = time.time()
            if N == 'full':
                eigvals, eigvecs = self.eig(i)
            else:
                eigvals, eigvecs = self.eigs(N,i)
            t1 = time.time()
            self.eigenvalues.append(eigvals)
            self.eigenvectors.append(eigvecs)
            self.time_to_find_eigensystem.append(t1-t0)

        # Subtract out the zero point energy
        self.ground_ZPE = np.min(self.eigenvalues[0])
        for i in range(len(self.eigenvalues)):
            self.eigenvalues[i] -= self.ground_ZPE
            
        # The RWA, using the ground electronic ground vibrational to the
        # lowest singly-excited vibronic transition
        self.ground_to_excited_transition = np.min(self.eigenvalues[1])
        for i in range(1,len(self.energies)):
            self.eigenvalues[i] -= i * self.ground_to_excited_transition

    def generate_eigenvalues_and_vectors_eig(self):
        """Use the methods from PolymerVibrations to construct the 
            eigenvalues and eigenmatrices for the GSM, SEM, and if 
            applicable, the DEM
"""
        self.eigenvalues = []
        self.eigenvectors = []

        N = self.params['number eigenvalues']

        for i in range(len(self.energies)):
            cut_off = N*len(self.energies[i])
            eigvals, eigvecs = self.eig(i)
            eigvals = eigvals[:cut_off]
            
            eigvecs = eigvecs[:,:cut_off]
            self.eigenvalues.append(eigvals)
            self.eigenvectors.append(eigvecs)

        # Subtract out the zero point energy
        self.ground_ZPE = np.min(self.eigenvalues[0])
        for i in range(len(self.eigenvalues)):
            self.eigenvalues[i] -= self.ground_ZPE
            
        # The RWA, using the ground electronic ground vibrational to the
        # lowest singly-excited vibronic transition
        self.ground_to_excited_transition = np.min(self.eigenvalues[1])
        for i in range(1,len(self.energies)):
            self.eigenvalues[i] -= i * self.ground_to_excited_transition

    def save_eigenvalues_and_vectors(self):
        eigval_save_name = os.path.join(self.base_path,'eigenvalues.npz')
        eigvec_save_name = os.path.join(self.base_path,'eigenvectors.npz')
        eigval_dict = {key:arr for key,arr in zip(self.manifolds,self.eigenvalues)}
        eigvec_dict = {key:arr for key,arr in zip(self.manifolds,self.eigenvectors)}
        np.savez(eigval_save_name,**eigval_dict)
        np.savez_compressed(eigvec_save_name,**eigvec_dict)
        eigen_params = {'final truncation size':self.truncation_size,
                        'ground to excited transition':float(self.ground_to_excited_transition),
                        'ground zero point energy':float(self.ground_ZPE),
                        'diagonalization time':self.time_to_find_eigensystem}
        with open(os.path.join(self.base_path,'eigen_params.yaml'),'w+') as yamlstream:
            yamlstream.write( yaml.dump(eigen_params,
                                        default_flow_style=False))

    def load_eigenvalues_and_vectors(self):
        eigval_save_name = os.path.join(self.base_path,'eigenvalues.npz')
        eigvec_save_name = os.path.join(self.base_path,'eigenvectors.npz')
        eigval_archive = np.load(eigval_save_name)
        eigvec_archive = np.load(eigvec_save_name)
        self.eigenvalues = [eigval_archive[key] for key in self.manifolds]
        self.eigenvectors = [eigvec_archive[key] for key in self.manifolds]

        with open(os.path.join(self.base_path,'eigen_params.yaml'),'r') as yamlstream:
            eigen_params = yaml.load(yamlstream,Loader=yaml.SafeLoader)
            self.truncation_size = eigen_params['final truncation size']
            self.ground_ZPE = eigen_params['ground zero point energy']
            self.ground_to_excited_transition = eigen_params['ground to excited transition']
            try:
                self.time_to_find_eigensystem = eigen_params['diagonalization time']
            except KeyError:
                pass

    def electronic_trace(self,manifold_num,eignum):
        """Traces out the vibrational space, and returns the partial density matrix
in the electronic site basis"""
        num_sites = len(self.energies[manifold_num])
        vibration_space_size = self.truncation_size**self.num_vibrations
        # eigvec = self.eigenvectors[manifold_num][:,eignum].toarray().reshape((num_sites,vibration_space_size))
        eigvec = self.eigenvectors[manifold_num][:,eignum].reshape((num_sites,vibration_space_size))
        
        partial_trace = np.zeros((vibration_space_size,vibration_space_size),dtype='complex')
        num_sites = len(self.energies[manifold_num])
        for i in range(num_sites):
            vec = eigvec[i,...]
            partial_trace += np.outer(vec,vec)
        return partial_trace

    def vibrational_trace(self,manifold_num,eignum):
        """Traces out the vibrational space, and returns the partial density matrix
in the electronic site basis"""
        num_sites = len(self.energies[manifold_num])
        vibration_space_size = self.truncation_size**self.num_vibrations
        # eigvec = self.eigenvectors[manifold_num][:,eignum].toarray().reshape((num_sites,vibration_space_size))
        eigvec = self.eigenvectors[manifold_num][:,eignum].reshape((num_sites,vibration_space_size))
        
        partial_trace = np.zeros((num_sites,num_sites),dtype='complex')
        for i in range(num_sites):
            for j in range(num_sites):
                vec_i = eigvec[i,...]
                vec_j = eigvec[j,...]
                rho_ij = np.outer(vec_i,vec_j)
                partial_trace[i,j] = rho_ij.trace()
        return partial_trace
    
    def trace_norm_diff(self,eignum1,eignum2,*,manifold_num=1,partial_trace = 'vibrational'):
        if partial_trace == 'vibrational':
            rho1 = self.vibrational_trace(manifold_num,eignum1)
            rho2 = self.vibrational_trace(manifold_num,eignum2)
        elif partial_trace == 'electronic':
            rho1 = self.electronic_trace(manifold_num,eignum1)
            rho2 = self.electronic_trace(manifold_num,eignum2)
        diff = rho1 - rho2
        sing_vals = np.linalg.svd(diff,compute_uv=False)
        return np.sum(sing_vals)/2

    def dodonov_distance(self,eignum1,eignum2,*,manifold_num=1,partial_trace = 'vibrational'):
        if partial_trace == 'vibrational':
            rho1 = self.vibrational_trace(manifold_num,eignum1)
            rho2 = self.vibrational_trace(manifold_num,eignum2)
        elif partial_trace == 'electronic':
            rho1 = self.electronic_trace(manifold_num,eignum1)
            rho2 = self.electronic_trace(manifold_num,eignum2)
        purity1 = np.dot(rho1,rho1).trace()
        purity2 = np.dot(rho2,rho2).trace()
        overlap = np.dot(rho1,rho2).trace()/np.sqrt(purity1*purity2)
        return 1 - np.abs(overlap)



#Standard python libraries
import numpy as np
import os
import yaml
import warnings
import numbers
import itertools
from datetime import datetime
import time
import matplotlib.pyplot as plt
from scipy.sparse import csr_matrix

from .eigenstates import PolymerVibrations
from .base_class import DataOrganizer

class SimpleEigenGenerator(PolymerVibrations,DataOrganizer):
    """This class generates the eigenvalues and eigenvectors
        needed for the specified system in the params file.  This
        class should be initialized once at the beginning of any calculation 
        run, and the generate and save methods should both be called once."""
    def __init__(self,parameter_file_path,*,
                 qr_flag=False,mask_by_occupation_num=True):
        parameter_file = os.path.join(parameter_file_path,'params.yaml')
        super().__init__(parameter_file,qr_flag=qr_flag,
                         mask_by_occupation_num=mask_by_occupation_num)
        self.base_path = parameter_file_path
        self.load_params()
        self.set_diagrams_and_manifolds()
        self.generate_eigenvalues_and_vectors_low_memory()
        self.combine_eigensystem()

    def set_diagrams_and_manifolds(self):
        if len(self.energies) == 2:
            self.diagrams = ['GSB', 'SE']
            self.manifolds = ['GSM','SEM']
            self.diagrams_by_manifold = {'GSM':['GSB'],'SEM':['SE']}
        elif len(self.energies) == 3:
            self.diagrams = ['GSB','SE','ESA']
            self.manifolds = ['GSM','SEM','DEM']
            self.diagrams_by_manifold = {'GSM':['GSB'],'SEM':['SE','ESA']}

    def generate_eigenvalues_and_vectors_low_memory(self):
        """Use the methods from PolymerVibrations to construct the 
            eigenvalues and eigenmatrices for the GSM, SEM, and if 
            applicable, the DEM
"""
        self.time_to_find_eigensystem = []

        N = self.params['number eigenvalues']

        for i in range(len(self.energies)):
            t0 = time.time()
            if N == 'full':
                print('diagonalizing 1 manifold...')
                eigvals, eigvecs = self.eig(i)
            else:
                eigvals, eigvecs = self.eigs(N,i)
            t1 = time.time()
            np.save(os.path.join(self.base_path,'Manifold{}_eigenvalues.npy'.format(i)),eigvals)
            np.save(os.path.join(self.base_path,'Manifold{}_eigenvectors.npy'.format(i)),eigvecs)
            del eigvals
            del eigvecs
            self.time_to_find_eigensystem.append(t1-t0)
        times = np.array([self.time_to_find_eigensystem])
        np.save(os.path.join(self.base_path,'Times_to_diagonalize.npy'),times)

    def combine_eigensystem(self):
        self.eigenvalues = []
        self.eigenvectors = []
        for i in range(len(self.energies)):
            eigvals = np.load(os.path.join(self.base_path,'Manifold{}_eigenvalues.npy'.format(i)))
            eigvecs = np.load(os.path.join(self.base_path,'Manifold{}_eigenvectors.npy'.format(i)))
            self.eigenvalues.append(eigvals)
            self.eigenvectors.append(eigvecs)
        try:
            times = self.time_to_find_eigensystem
        except:
            times = np.load(os.path.join(self.base_path,'Times_to_diagonalize.npy'))
            times = [float(time) for time in times]

        self.ground_ZPE = np.min(self.eigenvalues[0])
        for i in range(len(self.eigenvalues)):
            self.eigenvalues[i] -= self.ground_ZPE
            
        # The RWA, using the ground electronic ground vibrational to the
        # lowest singly-excited vibronic transition
        self.ground_to_excited_transition = np.min(self.eigenvalues[1])
        for i in range(1,len(self.energies)):
            self.eigenvalues[i] -= i * self.ground_to_excited_transition

        eigval_save_name = os.path.join(self.base_path,'eigenvalues.npz')
        eigvec_save_name = os.path.join(self.base_path,'eigenvectors.npz')
        eigval_dict = {key:arr for key,arr in zip(self.manifolds,self.eigenvalues)}
        eigvec_dict = {key:arr for key,arr in zip(self.manifolds,self.eigenvectors)}
        np.savez(eigval_save_name,**eigval_dict)
        np.savez_compressed(eigvec_save_name,**eigvec_dict)
        eigen_params = {'final truncation size':self.truncation_size,
                        'ground to excited transition':float(self.ground_to_excited_transition),
                        'ground zero point energy':float(self.ground_ZPE),
                        'diagonalization time':times}
        with open(os.path.join(self.base_path,'eigen_params.yaml'),'w+') as yamlstream:
            yamlstream.write( yaml.dump(eigen_params,
                                        default_flow_style=False))




from .eigen_generator import EigenGenerator
from .simple_eigen_generator import SimpleEigenGenerator
from .dipole_operator import DipoleConverter
from .params_converter import convert
from .dipole_operator import CalculateCartesianDipoleOperator

from .Hamiltonians import LadderOperators, DisplacedAnharmonicOscillator, PolymerVibrations,DiagonalizeHamiltonian

from .DiabaticLindblad_Liouvillians import OpenPolymer, OpenPolymerVibrations
from .Hamiltonians import PolymerVibrations,DiagonalizeHamiltonian
from .Redfield_Liouvillians import RedfieldConstructor,DiagonalizeLiouvillian,SecularRedfieldConstructor
from .manual_L_input import ManualL
from .run_HLG import run



#Standard python libraries
import os
import warnings
import copy
import time
import itertools
import functools

#Dependencies - numpy, scipy, matplotlib, pyfftw
import numpy as np
import pyx

class DiagramDrawer:
    """This class is used to draw double-sided Feynman diagrams and save them as pdf files
"""
    def __init__(self):
        self.draw_functions = {'Ku':self.draw_Ku,'Kd':self.draw_Kd,'Bu':self.draw_Bu,'Bd':self.draw_Bd}
        self.pulse_labels = ['a','b','c','d','e','f','g','h','i','j','k','l','m',
                             'n','o','p','q','r','s','t','u','v','w','x','y','z']

    def save_diagram(self,diagram,*,folder_name = ''):
        os.makedirs(folder_name,exist_ok=True)
        self.c = pyx.canvas.canvas()
        filename = ''
        interaction_counter = 0
        for KB,n in diagram:
            self.draw_functions[KB](n,interaction_counter)
            filename += KB + str(n)
            interaction_counter += 1
        self.c.writePDFfile(os.path.join(folder_name,filename))

    def save_diagrams(self,diagrams,*,folder_name=''):
        for diagram in diagrams:
            self.save_diagram(diagram,folder_name=folder_name)

    def display_diagram(self,diagram,*,exclude="image/png"):
        """Displays a diagram in a Jupyter notebook environment or similar
        Args:
            diagram: Feynman diagram to be drawn, must be a list or tuple of tuples
            exclude: MIME types to exclude when attempting to display diagram 
"""
        self.c = pyx.canvas.canvas()
        interaction_counter = 0
        for KB,n in diagram:
            self.draw_functions[KB](n,interaction_counter)
            interaction_counter += 1
        display(self.c,exclude=exclude)

    def display_diagrams(self,diagrams,*,exclude="image/png"):
        for diagram in diagrams:
            self.display_diagram(diagram,exclude=exclude)

    def double_sided(self,pulse_num):
        self.c.stroke(pyx.path.line(0,pulse_num,0,pulse_num+1))
        self.c.stroke(pyx.path.line(1, pulse_num, 1, pulse_num+1))

    def K_circle(self,pulse_num):
        self.c.fill(pyx.path.circle(0,pulse_num+0.5,0.1))

    def B_circle(self,pulse_num):
        self.c.fill(pyx.path.circle(1,pulse_num+0.5,0.1))

    def right_arrow(self,x,y,pulse_num):
        xf, yf = (x+0.5,y+0.5)
        self.c.stroke(pyx.path.line(x,y,xf,yf),[pyx.deco.earrow(size=0.3)])
        return xf,yf

    def left_arrow(self,x,y,pulse_num):
        xf,yf = (x-.5,y+.5)
        self.c.stroke(pyx.path.line(x,y,xf,yf),[pyx.deco.earrow(size=0.3)])
        return xf,yf

    def pulse_text(self,x,y,pulse_num):
        text = self.pulse_labels[pulse_num]
        self.c.text(x,y,text)

    def draw_Bd(self,pulse_num,interaction_num):
        m = interaction_num
        n = pulse_num
        self.double_sided(m)
        self.B_circle(m)
        x,y = (1,m+0.5)
        xf,yf = self.right_arrow(x,y,m)
        self.pulse_text(xf-0.1,yf-0.5,n)

    def draw_Bu(self,pulse_num,interaction_num):
        m = interaction_num
        n = pulse_num
        self.double_sided(m)
        self.B_circle(m)
        x,y = (1.5,m)
        self.left_arrow(x,y,m)
        self.pulse_text(x-0.1,y+0.1,n)

    def draw_Kd(self,pulse_num,interaction_num):
        m = interaction_num
        n = pulse_num
        self.double_sided(m)
        self.K_circle(m)
        xi,yi = (0,m+0.5)
        xf,yf = self.left_arrow(xi,yi,m)
        self.pulse_text(xf-0.1,yf-0.5,n)

    def draw_Ku(self,pulse_num,interaction_num):
        m = interaction_num
        n = pulse_num
        self.double_sided(m)
        self.K_circle(m)
        xi,yi = (-0.5,m)
        xf,yf = self.right_arrow(xi,yi,m)
        self.pulse_text(xi-0.1,yi+0.1,n)

    

class DiagramGenerator(DiagramDrawer):
    """

    Args:
        detection_type (str): default is 'polarization', other options are 'integrated_polarization' or 'fluorescence'

"""
    def __init__(self,*,detection_type = 'polarization'):
        DiagramDrawer.__init__(self)

        # Code will not actually function until the following three empty
        # lists are set by the user
        self.efield_times = [] #initialize empty list of times assoicated with each electric field shape
        self.efield_wavevectors = []

        # Change this, if applicable, to the maximum number of manifolds in the system under study
        self.maximum_manifold = np.inf

        # Change this to a negative number, possibly -np.inf, if the initial state can be de-excited
        self.minimum_manifold = 0

        # Used for automatically generating diagram instructions
        self.wavevector_dict = {'-':('Bu','Kd'),'+':('Ku','Bd')}

        # Used to find resonant-only contributions
        self.instruction_to_manifold_transition = {'Bu':np.array([0,1]),
                                                   'Bd':np.array([0,-1]),
                                                   'Ku':np.array([1,0]),
                                                   'Kd':np.array([-1,0])}
        self.detection_type = detection_type
        
        if detection_type == 'polarization' or detection_type == 'integrated_polarization':
            self.filter_instructions = self.polarization_detection_filter_instructions
        elif detection_type == 'fluorescence':
            self.filter_instructions = self.fluorescence_detection_filter_instructions

    def interaction_tuple_to_str(self,tup):
        """Converts a tuple, tup = (nr,nc) into a string of +'s and -'s
"""
        s = '+'*tup[0] + '-'*tup[1]
        return s

    def set_phase_discrimination(self,interaction_list):
        if type(interaction_list[0]) is str:
            new_list = interaction_list
        elif type(interaction_list[0]) is tuple:
            new_list = [self.interaction_tuple_to_str(el) for el in interaction_list]

        self.efield_wavevectors = new_list
        self.set_pdc()

        # If pulses and/or phase discrimination are changed, these two attributes 
        # must be reset or removed:
        try:
            del self.pulse_sequence
            del self.pulse_overlap_array
        except AttributeError:
            pass

    def set_pdc(self):
        num_pulses = len(self.efield_wavevectors)
        pdc = np.zeros((num_pulses,2),dtype='int')
        for i in range(num_pulses):
            for j in range(len(self.efield_wavevectors[i])):
                if self.efield_wavevectors[i][j] == '+':
                    pdc[i,0] += 1
                elif self.efield_wavevectors[i][j] == '-':
                    pdc[i,1] += 1
                else:
                    raise Exception('Could not set phase-discrimination condition')
        self.pdc = pdc
        self.pdc_tuple = tuple(tuple(pdc[i,:]) for i in range(pdc.shape[0]))

    def save_diagrams_as_text(self,diagrams,*,file_name = 'auto'):
        if file_name == 'auto':
            file_name = ''
            for nr,nc in self.pdc_tuple:
                file_name += '{1}_{2}_'.format(nr,nc)
            file_name = file_name[:-1]
            file_name += '.txt'
        with open(file_name,'w') as txtfile:
            txtfile.writelines('['+str(diagrams[0]))
            for diagram in diagrams[1:]:
                txtfile.writelines(',\n '+str(diagram))
            txtfile.writelines(']')

    def polarization_detection_filter_instructions(self,instructions):
        rho_manifold = np.array([0,0])
        for ins in instructions:
            rho_manifold += self.instruction_to_manifold_transition[ins]
            if rho_manifold[0] < self.minimum_manifold or rho_manifold[1] < self.minimum_manifold:
                return False
            if rho_manifold[0] > self.maximum_manifold or rho_manifold[1] > self.maximum_manifold:
                return False
        if rho_manifold[0] - rho_manifold[1] == 1:
            return True
        else:
            return False

    def fluorescence_detection_filter_instructions(self,instructions):
        rho_manifold = np.array([0,0])
        for ins in instructions:
            rho_manifold += self.instruction_to_manifold_transition[ins]
            if rho_manifold[0] < self.minimum_manifold or rho_manifold[1] < self.minimum_manifold:
                return False
            if rho_manifold[0] > self.maximum_manifold or rho_manifold[1] > self.maximum_manifold:
                return False
        if abs(rho_manifold[1]-rho_manifold[1]) == 0 and rho_manifold[1] != 0:
            return True
        else:
            return False

    def instructions_from_permutation(self,perm):
        f_list = []
        efield_order = []
        for i,k in perm:
            f_list.append(self.wavevector_dict[k])
            efield_order.append(i)
        all_instructions = itertools.product(*f_list)
        filtered_instructions = []
        for ins in all_instructions:
            if self.filter_instructions(ins):
                filtered_instructions.append(tuple(zip(ins,efield_order)))
        return filtered_instructions

    def wavefunction_instructions_from_permutation(self,perm):
        rho_instructions = self.instructions_from_permutation(perm)
        psi_instructions = []
        for instructions in rho_instructions:
            new_instructions = self.convert_rho_instructions_to_psi_instructions(instructions)
            if new_instructions in psi_instructions:
                pass
            else:
                psi_instructions.append(new_instructions)

        return psi_instructions

    def convert_rho_instructions_to_psi_instructions(self,instructions):
        psi_instructions = {'ket':[],'bra':[]}
        for key, pulse_num in instructions:
            if key[0] == 'K':
                psi_instructions['ket'].append((key[1],pulse_num))
            elif key[0] == 'B':
                psi_instructions['bra'].append((key[1],pulse_num))
        return psi_instructions

    def convert_rho_instructions_list_to_psi_instructions_list(self,instructions_list):
        psi_instructions_list = []
        for instructions in instructions_list:
            psi_instructions = self.convert_rho_instructions_to_psi_instructions(instructions)
            if psi_instructions not in psi_instructions_list:
                psi_instructions_list.append(psi_instructions)
        return psi_instructions_list

    def relevant_permutations(self,pulse_time_meshes):
        self.set_ordered_interactions()
        all_permutations = set(itertools.permutations(self.ordered_interactions))
        filtered_permutations = []
            
        for perm in all_permutations:
            remove = False
            for i in range(len(perm)-1):
                indi = perm[i][0]
                ti = pulse_time_meshes[indi]
                for j in range(i+1,len(perm)):
                    indj = perm[j][0]
                    tj = pulse_time_meshes[indj]
                    if ti[0] > tj[-1]:
                        remove = True
                        break
                if remove == True:
                    break
            if not remove:
                filtered_permutations.append(perm)
        return filtered_permutations

    def set_ordered_interactions(self):
        """Sets up a list of the time-ordered interactions, along with associated wavevector ('+' or '-')
"""
        num_pulses = len(self.efield_wavevectors)
        self.ordered_interactions = []
        for i in range(num_pulses):
            ks = self.efield_wavevectors[i]
            for k in ks:
                self.ordered_interactions.append((i,k))

    def check_new_diagram_conditions(self,arrival_times):
        new = np.array(arrival_times)
        new_pulse_sequence = np.argsort(new)
        new_pulse_overlap_array = np.ones((len(arrival_times),
                                            len(arrival_times)),
                                           dtype='bool')

        intervals = self.arrival_times_to_pulse_intervals(arrival_times)
        
        for i in range(len(intervals)):
            ti = intervals[i]
            for j in range(i+1,len(intervals)):
                tj = intervals[j]
                if ti[0] > tj[-1]:
                    new_pulse_overlap_array[i,j] = False
                elif ti[-1] < tj[0]:
                    new_pulse_overlap_array[i,j] = False
        try:
            logic_statement = (np.allclose(new_pulse_overlap_array,self.pulse_overlap_array)
                and np.allclose(new_pulse_sequence,self.pulse_sequence))
            if logic_statement:
                calculate_new_diagrams = False
            else:
                calculate_new_diagrams = True
        except AttributeError:
            calculate_new_diagrams = True
        
        self.pulse_sequence = new_pulse_sequence
        self.pulse_overlap_array = new_pulse_overlap_array

        return calculate_new_diagrams

    def arrival_times_to_pulse_intervals(self,arrival_times):
        if self.detection_type == 'polarization' or self.detection_type == 'integrated_polarization':
            if len(arrival_times) == len(self.efield_wavevectors) + 1:
                # If the arrival time of the local oscillator was included in the list arrival_times,
                # remove it, it is not relevant to diagram generation
                arrival_times = arrival_times[:-1]
        intervals = [self.efield_times[i] + arrival_times[i] for i in range(len(arrival_times))]

        return intervals

    def set_diagrams(self,arrival_times):
        intervals = self.arrival_times_to_pulse_intervals(arrival_times)
        
        efield_permutations = self.relevant_permutations(intervals)
        all_instructions = []
        for perm in efield_permutations:
            all_instructions += self.instructions_from_permutation(perm)
        self.current_diagrams = all_instructions
        

    def get_diagrams(self,arrival_times):
        calculate_new_diagrams = self.check_new_diagram_conditions(arrival_times)
        if calculate_new_diagrams:
            self.set_diagrams(arrival_times)
        return self.current_diagrams

    def get_wavefunction_diagrams(self,arrival_times):
        rho_instructions_list = self.get_diagrams(arrival_times)
        wavefunction_instructions = self.convert_rho_instructions_list_to_psi_instructions_list(rho_instructions_list)
        return wavefunction_instructions

    def get_diagram_final_state(self,diagram):
        """Returns ket and bra manifold indices after all diagram interactions
        Args:
            diagram (list) : list of ordered interactions
        Returns: list with the first entry the ket manifold index, and 
            the second entry the bra manifold index
"""
        rho_manifold = np.array([0,0],dtype='int')
        for ins,pulse_num in diagram:
            rho_manifold += self.instruction_to_manifold_transition[ins]
        return list(rho_manifold)

    def get_diagram_excitation_manifold(self,diagram,*,number_of_interactions=2):
        rho_manifold = np.array([0,0])
        for ins,pulse_num in diagram[:number_of_interactions]:
            rho_manifold += self.instruction_to_manifold_transition[ins]
        if rho_manifold[0] == rho_manifold[1]:
            return rho_manifold[0]
        else:
            return None

    def filter_diagrams_by_final_state(self,diagrams,state):
        """Returns all diagrams that end in the specified state
        Args:
            diagrams (list) : list of diagrams to filter
            state (list) : list with the first entry the ket manifold index, 
                and the second entry the bra manifold index
"""
        new_diagrams = []
        for diagram in diagrams:
            diagram_state = self.get_diagram_final_state(diagram)
            if diagram_state == state:
                new_diagrams.append(diagram)
        return new_diagrams

    def filter_diagrams_by_excitation_manifold(self,diagrams,*,manifold=1,number_of_interactions=2):
        new_diagrams = []
        for diagram in diagrams:
            man = self.get_diagram_excitation_manifold(diagram,number_of_interactions=number_of_interactions)
            if man == manifold:
                new_diagrams.append(diagram)
        return new_diagrams

    def get_diagram_sign(self,diagram):
        if len(diagram) % 2:
            sign = 1j
        else:
            sign = 1
        for ins,pulse_num in diagram:
            if 'K' in ins:
                sign *= -1j
            elif 'B' in ins:
                sign *= 1j
            else:
                raise Exception('diagram is not in proper format')
            
        return sign

    def filter_diagrams_by_sign(self,diagrams,*,sign=1):
        new_diagrams = []
        for diagram in diagrams:
            diagram_sign = self.get_diagram_sign(diagram)
            if diagram_sign == sign:
                new_diagrams.append(diagram)
        return new_diagrams

    def remove_all_permutations(self,diagrams):
        new_diagrams = []
        diagram_weights = []
        set_list = []
        for diagram in diagrams:
            new_set = set(diagram)
            try:
                ind = set_list.index(new_set)
                diagram_weights[ind] += 1
            except ValueError:
                new_diagrams.append(diagram)
                diagram_weights.append(1)
                set_list.append(new_set)

        return diagram_weights, new_diagrams



import numpy as np
import matplotlib.pyplot as plt
from matplotlib.widgets import Slider
import matplotlib
from scipy.sparse import csr_matrix, identity, kron
from scipy.sparse.linalg import eigs, eigsh
import itertools
from scipy.linalg import block_diag, eig, expm, eigh
import yaml
import copy
from scipy.special import factorial, binom
from numpy.polynomial.hermite import hermval
import warnings

class LadderOperators:
    def __init__(self,size):
        self.size = size
        self.calculation_size = size + 20
        self.set_x_and_p()

    def set_x_and_p(self):
        self.x = (self.create() + self.destroy())/np.sqrt(2)
        self.p = 1j* (-self.create() + self.destroy())/np.sqrt(2)

    def create(self):
        def offdiag1(n):
            return np.sqrt(n+1)

        n = np.arange(0,self.calculation_size)
        off1 = offdiag1(n[0:-1])
        ham = np.zeros((self.calculation_size,self.calculation_size))
        ham += np.diag(off1,k=-1)
        return ham

    def destroy(self):
        def offdiag1(n):
            return np.sqrt(n+1)

        n = np.arange(0,self.calculation_size)
        off1 = offdiag1(n[0:-1])
        ham = np.zeros((self.calculation_size,self.calculation_size))
        ham += np.diag(off1,k=1)
        return ham

    def x_power_n(self,n):
        xn = np.linalg.matrix_power(self.x,n)
        return xn[:self.size,:self.size]

    def p_power_n(self,n):
        pn = np.linalg.matrix_power(self.p,n)
        return pn[:self.size,:self.size]
    
class ArbitraryHamiltonian(LadderOperators):
    def set_ham(self,constant,pcoef,xcoef,*,real=True):
        self.constant = constant
        self.pcoef = pcoef
        self.xcoef = xcoef
        ham = np.diag(np.ones(self.calculation_size,dtype=complex)*constant)
        for n in range(len(pcoef)):
            ham += pcoef[n] * np.linalg.matrix_power(self.p,n+1)
        for m in range(len(xcoef)):
            ham += xcoef[m] * np.linalg.matrix_power(self.x,m+1)
        self.ham = ham[:self.size,:self.size]
        if real:
            self.ham = np.real(self.ham)

    def set_eigs(self,num,*,hermitian=True):
        if hermitian:
            eigvals, eigvecs = eigsh(self.ham,k=num,which='SM')
        else:
            eigvals, eigvecs = eigs(self.ham,k=num,which='SM')
        sort_indices = eigvals.argsort()
        eigvals.sort()
        eigvecs = eigvecs[:,sort_indices]
        # I choose to pick the phase of my eigenvectors such that the state which has the
        # largest overlap has a positive overlap. For sufficiently small d, and alpha close
        # to 1, this will be the overlap between the same excited and ground states.
        for i in range(num):
            max_index = np.argmax(np.abs(eigvecs[:,i]))
            if eigvecs[max_index,i] < 0:
                eigvecs[:,i] *= -1
        self.eigvecs = eigvecs
        if np.max(np.imag(eigvals)) > 1E-12:
            warnings.warn('Eigenvalues are not fully real')
            self.eigvals = eigvals
        else:
            self.eigvals = np.real(eigvals)

    def potential(self,x):
        y = np.ones(x.shape,x.dtype)*self.constant
        for i in range(len(self.xcoef)):
            y += np.power(x,i+1) * self.xcoef[i]
        return y

    def plot(self,*,x=np.arange(-3,3,.01)):
        fig, ax = plt.subplots()
        ax.plot(x,self.potential(x))
        for n in range(len(self.eigvals)):
            ev = self.eigvals[n]
            vec = self.eigvecs[:,n]
            ax.axhline(y=ev,linestyle='--')
            f = oscillator_1d(vec,x)
            ax.plot(x,f**2+ev)
            # ax.plot(x,np.real(f)+ev,x,np.imag(f)+ev)

    def displaced_vacuum(self,*,disp = 1):
        vac = np.zeros(self.calculation_size)
        vac[0] = 1
        disp_oper = expm(disp * self.create() + np.conjugate(disp) * self.destroy())
        disp_vac = disp_oper.dot(vac)
        disp_vac = disp_vac[:len(self.eigvals)]
        return disp_vac

    def time_evolution(self,state,time):
        basis_change = self.eigvecs[:len(self.eigvals),:len(self.eigvals)]
        state = np.conjugate(basis_change.T).dot(state)
        time_oper = np.exp(1j * self.eigvals * time)
        state = state * time_oper
        state = basis_change.dot(state)
        return state

    def animate_wavepacket(self,*,disp = 1,x=np.arange(-3,3,0.01),t=np.arange(0,2*np.pi,np.pi/30)):
        disp_vac = self.displaced_vacuum(disp = disp)
        fig, ax = plt.subplots()
        ax.plot(x,self.potential(x))
        fig.subplots_adjust(bottom=0.4)
        t_ax = fig.add_axes([0.1,0.05,0.65,0.02])
        t_slider = Slider(t_ax, 'time', 0, t[-1],valinit=0)

        psi = oscillator_1d(disp_vac,x)

        line_r, line_i, line_abs, = ax.plot(x,np.real(psi)*2+2,
                                            x,np.imag(psi)*2+2,
                                            x,np.abs(psi)*2+2)
        
        def update(val):
            t = t_slider.val
            psi = oscillator_1d(self.time_evolution(disp_vac,t),x)
            line_r.set_ydata(np.real(psi)*2+2)
            line_i.set_ydata(np.imag(psi)*2+2)
            line_abs.set_ydata(np.abs(psi)*2+2)
        t_slider.on_changed(update)

class AnharmonicDisplaced(ArbitraryHamiltonian):
    """Implements displacement in x for all powers. Does not explicitly treat a linear coupling.
        This class can model all the same things that ArbitraryHamiltonian can.  However, it is
        much easier to work with when you need to displace an anharmonic oscillator in x
"""
    def __init__(self,size):
        """
        Args:
            size (int) : truncation size for Hamiltonian
"""
        self.size = size

    def set_ham(self,constant,displacement,pcoef,xcoef,*,real=True):
        """Creates anharmonic Hamiltonian for 1 vibrational mode
        Args:
            constant (complex) : complex reorganization energy
            displacement (complex) : phase-space displacement
            pcoef (list) : coefficients for 2nd order and higher momentum terms
            xcoef (list) : coefficients for 2nd order and higher position terms
"""
        self.constant = constant
        self.p0 = np.imag(displacement)
        self.x0 = np.real(displacement)
        self.pcoef = pcoef
        self.xcoef = xcoef
        extra_size = max(len(pcoef),len(xcoef))
        # Calculates powers of x and p via matrix powers. Must use an increased
        # base truncation size in order to correctly resolve x^n and p^n
        self.calculation_size = self.size + extra_size
        self.set_x_and_p()
        ham = np.diag(np.ones(self.calculation_size,dtype=complex)*constant)
        for n in range(0,len(pcoef)):
            p = self.p - np.diag(np.ones(self.calculation_size))*self.p0
            ham += pcoef[n] * np.linalg.matrix_power(p,n+2)
        for m in range(0,len(xcoef)):
            x = self.x - np.diag(np.ones(self.calculation_size))*self.x0
            ham += xcoef[m] * np.linalg.matrix_power(x,m+2)
        self.ham = ham[:self.size,:self.size]
        if real:
            self.ham = np.real(self.ham)

    def potential(self,x):
        y = np.ones(x.shape,x.dtype)*self.constant
        for i in range(len(self.xcoef)):
            y += np.power(x-self.x0,i+2) * self.xcoef[i]
        return y

    def plot(self,*,x=np.arange(-3,3,.01)):
        fig, ax = plt.subplots()
        ax.plot(x,self.potential(x))
        for n in range(len(self.eigvals)):
            ev = self.eigvals[n]
            vec = self.eigvecs[:,n]
            ax.axhline(y=ev,linestyle='--')
            f = oscillator_1d(vec,x)
            ax.plot(x,f**2+ev)
    
class Polymer:
    """This class creates a matrix for an electronic polymer in site basis.
It is hard-coded to always set the ground-state energy to 0.  This version 
requires more inputs, and does not do additional calculations, as compared
to PolymerOld"""
    def __init__(self,site_energies,site_couplings,*,auto_DEM=True):
        """This initializes an object with an arbitrary number of site
energies and couplings

site_energies - list of excitation energies of individual sites

site_couplings - list of energetic couplings between singly-excited electronic 
states in the site basis, for example [J12,J13,...,J1N,J23,...,J2N,...]
"""
        try:
            self.num_sites = len(site_energies[0])
            self.energies = site_energies
            self.couplings = site_couplings
        except TypeError:    
            self.num_sites = len(site_energies)
            self.energies = [[0.0],site_energies]
            self.couplings = [[],site_couplings]
        if self.num_sites > 1 and auto_DEM:
            self.couplings.append([])
            DEM_ham = self.electronic_manifold_2()
            self.energies.append(DEM_ham.diagonal())
        

    def electronic_manifold_0(self):
        return np.array([[0.0]])
    
    def electronic_manifold_1(self):
        """n-excited manifold"""
        J_iter = iter(self.couplings[1])
        ham = np.diag(np.array(self.energies[1],dtype=float))
        N = len(self.energies[1])
        for i in range(0,N-1):
            for j in range(i+1,N):
                ham[i,j] = next(J_iter)
                ham[j,i] = ham[i,j]
        return ham

    def electronic_manifold_2(self):
        SEM_ham = self.electronic_manifold_1()
        DEM_indices = list(itertools.combinations(range(len(self.energies[1])),2))
        DEM_energies = np.array([self.energies[1][i] + self.energies[1][j] for (i,j) in DEM_indices],dtype=float)
        DEM_ham = np.diag(DEM_energies)
        couplings_list = []
        for (i,j) in itertools.combinations(range(len(DEM_energies)),2):
            inds_i = set(DEM_indices[i])
            inds_j = set(DEM_indices[j])
            diff = tuple(inds_i ^ inds_j)
            if len(diff) == 2:
                DEM_ham[i,j] = SEM_ham[diff]
                DEM_ham[j,i] = DEM_ham[i,j]
                couplings_list.append(DEM_ham[i,j])
        self.couplings[2] = couplings_list
        return DEM_ham

    def electronic_manifold(self,n):
        if n == 0:
            return self.electronic_manifold_0()
        elif n == 1:
            return self.electronic_manifold_1()
        elif n == 2:
            return self.electronic_manifold_2()
        else:
            print('Cannot construct higher excited manifolds')
    
    def electronic_manifold_eig(self,n):
        ham = self.electronic_manifold(n)
        eigvals, eigvecs = eig(ham)
        eigvals = np.real(eigvals)
        sort_ind = eigvals.argsort()
        eigvals = eigvals[sort_ind]
        eigvecs = eigvecs[:,sort_ind]
        return eigvals, eigvecs
        
    def total_electronic_hamiltonian(self):
        """Only returns the GSM, SEM, and DEM"""
        return block_diag(*[self.electronic_manifold(n) for n in range(3)])

class PolymerVibrations(Polymer):
    def __init__(self,yaml_file,*,qr_flag=False,mask_by_occupation_num=False):
        """Initial set-up is the same as for the Polymer class, but I also need
to unpack the vibrational_frequencies, which must be passed as a nested list.
Each site may have N vibrational modes, and each has a frequency, a displacement
and a frequency shift for the excited state
for sites a, b, ...

qr_flag - defaults to False. If True, applies a qr factorization to each set of 
eigenvalues. This forces the eigenvalues to be orthogonal, in the case of
degenerate subspaces. If the hamiltonian is not hermitian, this will break
the eigenvalues (since they are not expected to be orthogonal).
"""
        with open(yaml_file) as yamlstream:
            params = yaml.load(yamlstream,Loader=yaml.SafeLoader)
        try:
            auto_DEM = params['auto_DEM']
        except KeyError:
            auto_DEM = True
        super().__init__(params['site_energies'],params['site_couplings'],auto_DEM=auto_DEM)
        self.truncation_size = params['initial truncation size']
        self.params = params
        self.occupation_num_mask = mask_by_occupation_num
        self.set_vibrations()
        self.qr_flag = False
        self.r_mats = []
        self.check_eigenvectors = False

    def vector_of_ones_kron(self,position,item):
        """Takes in a single vibrational hamiltonians and krons it with the correct 
            number of vibrational identities, inserting it into its position as indexed by its mode
            position as specified in the input file"""
        N = self.truncation_size
        nv = self.num_vibrations
        ones_list = [np.ones(N) for i in range(nv-1)]
        ones_list.insert(position,item)
        vec = ones_list.pop(0)
        for next_item in ones_list:
            vec = np.kron(vec,next_item)
        return vec

    def set_vibrational_total_occupation_number(self):
        N = self.truncation_size
        nv = self.num_vibrations
        single_mode_occ = np.arange(N)
        occ_num = self.vector_of_ones_kron(0,single_mode_occ)
        for i in range(1,nv):
            occ_num += self.vector_of_ones_kron(i,single_mode_occ)
        self.vibrational_total_occupation_number = occ_num

    def set_truncation_mask(self):
        """Creates a boolean mask to describe which states obey the truncation
           size collectively
"""
        N = self.truncation_size
        self.manifold_indices = []
        for i in range(len(self.energies)):
            num_excitations = len(self.energies[i])
            total_occ_num = np.kron(np.ones(num_excitations),self.vibrational_total_occupation_number)
            inds_to_keep = np.where(total_occ_num < N)
            self.manifold_indices.append(inds_to_keep)

    def set_vibrations(self):
        vibration_params = self.params['vibrations']
        # Vibrations in the ground manifold are assumed to be diagonal
        GSM_vib = [self.construct_vibrational_hamiltonians(mode_dict,0)
                           for mode_dict in vibration_params]
        
        self.num_vibrations = len(GSM_vib)
        
        # Vibrations in the SEM
        SEM_vib= [self.construct_vibrational_hamiltonians(mode_dict,1)
                           for mode_dict in vibration_params]

        self.vibrations = [GSM_vib,SEM_vib]
        
        # Vibrations in the DEM
        try:
            DEM_vib = [self.construct_vibrational_hamiltonians(mode_dict,2)
                           for mode_dict in vibration_params]
            self.vibrations.append(DEM_vib)
        except KeyError:
            pass

        if self.occupation_num_mask:
            self.set_vibrational_total_occupation_number()
            self.set_truncation_mask()
        
    def construct_vibrational_hamiltonians(self,single_mode,manifold_num):
        """For each vibrational mode, construct a list of sparse matrices defining the 
            vibrational hamiltonian for that mode in each excited state"""
        w = single_mode['omega_g']
        lams = single_mode['reorganization'+str(manifold_num)]
        ds  = single_mode['displacement'+str(manifold_num)]
        kins = single_mode['kinetic'+str(manifold_num)]
        pots = single_mode['potential'+str(manifold_num)]
        hams = []
        for l, d, kin, pot in zip(lams,ds,kins,pots):
            aho = AnharmonicDisplaced(self.truncation_size)
            aho.set_ham(l,d,kin,pot,real=True)
            hams.append(0.5 * w * aho.ham)
        return hams
    
    def site_basis(self,n,manifold_num):
        """This defines the site basis for the electronic space"""
        basis_matrix = np.zeros((len(self.energies[manifold_num]),len(self.energies[manifold_num])))
        basis_matrix[n,n] = 1
        return basis_matrix

    def single_site_vibrations(self,n,manifold_num):
        """This defines the hamiltonian on a single excited state given all vibrational modes"""
        vibrations_on_site = []
        for vibrational_mode in self.vibrations[manifold_num]:
            site_ham = vibrational_mode[n]
            vibrations_on_site.append(site_ham)
        total_ham = self.vibration_identity_kron(0,vibrations_on_site[0],manifold_num)
        for i in range(1,len(self.vibrations[manifold_num])):
            total_ham += self.vibration_identity_kron(i,vibrations_on_site[i],manifold_num)
        return kron(self.site_basis(n,manifold_num), total_ham)

    def linear_site_vibrational_couplings(self,mode_num,manifold_num):
        mode_dict = self.params['vibrations'][mode_num]
        coupling_list = mode_dict['linear_couplings'+str(manifold_num)]
        coupling_strength, site_index_pair = coupling_list[0]
        vib_coupling = self.single_linear_site_vibration_coupling(site_index_pair,mode_num,manifold_num)
        coupling_ham =  vib_coupling * coupling_strength
        
        for i in range(1,len(coupling_list)):
            coupling_strength, site_index_pair = coupling_list[i]
            vib_coupling = self.single_linear_site_vibration_coupling(site_index_pair,mode_num,manifold_num)
            coupling_ham +=  vib_coupling * coupling_strength
        return coupling_ham
    
    def single_linear_site_vibration_coupling(self,site_index_pair,mode_num,manifold_num):
        electronic_matrix = self.site_basis(0,manifold_num)
        electronic_matrix[:,:] = 0
        i,j = site_index_pair
        electronic_matrix[i,j] = 1
        electronic_matrix[j,i] = 1

        x = LadderOperators(self.truncation_size).x
        x = x[:self.truncation_size,:self.truncation_size]
        vib_coupling = self.vibration_identity_kron(mode_num,x,manifold_num)
        return kron(electronic_matrix,vib_coupling)
    
    def vibration_identity_kron(self,position,item,manifold_num):
        """Takes in a single vibrational hamiltonians and krons it with the correct 
            number of vibrational identities, inserting it into its position as indexed by its mode
            position as specified in the input file"""
        identities = [identity(self.truncation_size) for n in
                      range(len(self.vibrations[manifold_num])-1)]
        identities.insert(position,item)
        mat = identities.pop(0)
        for next_item in identities:
            mat = kron(mat,next_item)
        return mat
    
    def manifold_hamiltonian(self,manifold_num):
        """Creates requested hamiltonian
        Args:
            manifold_num (int): manifold - 0,1,2 correspond to GSM, SEM, DEM respectively
        Returns:
            (csr_matrix): hamiltonian of requested manifold as a sparse matrix
"""
        elec_manifold = self.electronic_manifold(manifold_num)
        vibrational_identity = identity(self.truncation_size)
        total_vibrational_identity = self.vibration_identity_kron(0,vibrational_identity,manifold_num)
        total_ham = kron(elec_manifold,total_vibrational_identity)
        for i in range(len(self.energies[manifold_num])):
            total_ham += self.single_site_vibrations(i,manifold_num)

        for i in range(self.num_vibrations):
            try:
                coupling_list = self.params['vibrations'][i]['linear_couplings'+str(manifold_num)]
                total_ham += self.linear_site_vibrational_couplings(i,manifold_num)
            except KeyError:
                pass
        if self.occupation_num_mask:
            inds = self.manifold_indices[manifold_num]
            total_ham = total_ham[inds[0]]
            total_ham = total_ham.transpose()
            total_ham = total_ham[inds[0]]
            total_ham = total_ham.transpose()
        
        return total_ham

    def eigs(self,num_eigvals,manifold_num):
        """Returns the requested number of eigvals, num_eigvals, sorted from smallest to largest,
            along with the corresponding eigenvectors reshaped into matrices with indices as follows:
            eigmat[electronic site, vibration 1, vibration 2, ..., eigen number], where eigen number is the
            index of the corresponding eigenvalue, starting from 0."""
        num_sites = len(self.energies[manifold_num])
        ham = self.manifold_hamiltonian(manifold_num)
        eigvals, eigvecs = eigsh(ham,k=num_eigvals*num_sites,which='SM')
        # Force degenerate eigenvectors to be orthogonal
        if self.qr_flag:
            eigvecs, r = np.linalg.qr(eigvecs,mode='reduced')
        if self.check_eigenvectors:
            HV = ham.dot(eigvecs)
            D = eigvecs.T.dot(HV)
            if np.allclose(D,np.diag(eigvals),rtol=1E-11,atol=1E-11):
                pass
            else:
                # warnings.warn('Eigenvalues altered by QR factorization, max absolute change in diagonal matrix of {}'.format(np.max(D-np.diag(eigvals))))
                warnings.warn('Using eigenvectors to diagonalize hamiltonian does not result in the expected diagonal matrix to tolerance, largest deviation is {}'.format(np.max(np.abs(D - np.diag(eigvals)))))
        
        sort_indices = eigvals.argsort()
        eigvals.sort()
        eigvecs = eigvecs[:,sort_indices]
        if self.qr_flag:
            r = r[:,sort_indices]
            self.r_mats.append(r)
        # I choose to pick the phase of my eigenvectors such that the state which has the
        # largest overlap has a positive overlap. For sufficiently small d, and alpha close
        # to 1, this will be the overlap between the same excited and ground states.
        for i in range(num_eigvals):
            max_index = np.argmax(np.abs(eigvecs[:,i]))
            if eigvecs[max_index,i] < 0:
                eigvecs[:,i] *= -1

        return eigvals, eigvecs

    def eig(self,manifold_num):
        """Returns the requested number of eigvals, num_eigvals, sorted from smallest to largest,
            along with the corresponding eigenvectors reshaped into matrices with indices as follows:
            eigmat[electronic site, vibration 1, vibration 2, ..., eigen number], where eigen number is the
            index of the corresponding eigenvalue, starting from 0."""
        num_sites = len(self.energies[manifold_num])
        ham = self.manifold_hamiltonian(manifold_num).toarray()
        eigvals, eigvecs = eigh(ham)
        # Force degenerate eigenvectors to be orthogonal
        if self.qr_flag:
            eigvecs, r = np.linalg.qr(eigvecs,mode='reduced')
        if self.check_eigenvectors:
            HV = ham.dot(eigvecs)
            D = eigvecs.T.dot(HV)
            if np.allclose(D,np.diag(eigvals),rtol=1E-11,atol=1E-11):
                pass
            else:
                # warnings.warn('Eigenvalues altered by QR factorization, max absolute change in diagonal matrix of {}'.format(np.max(D-np.diag(eigvals))))
                warnings.warn('Using eigenvectors to diagonalize hamiltonian does not result in the expected diagonal matrix to tolerance, largest deviation is {}'.format(np.max(np.abs(D - np.diag(eigvals)))))
        
        sort_indices = eigvals.argsort()
        eigvals.sort()
        eigvecs = eigvecs[:,sort_indices]
        if self.qr_flag:
            r = r[:,sort_indices]
            self.r_mats.append(r)
        # I choose to pick the phase of my eigenvectors such that the state which has the
        # largest overlap has a positive overlap. For sufficiently small d, and alpha close
        # to 1, this will be the overlap between the same excited and ground states.
        for i in range(eigvals.size):
            max_index = np.argmax(np.abs(eigvecs[:,i]))
            if eigvecs[max_index,i] < 0:
                eigvecs[:,i] *= -1

        return eigvals, eigvecs

    def serial_eigs(self,num_eigvals,manifold_num):
        vib_params = self.params['vibrations']
        N = len(vib_params) # number of vibrations
        omega_gs = np.array([vib_params[i]['omega_g'] for i in range(N)])
        diag_site_energies, electronic_eigvecs = self.electronic_manifold_eig(manifold_num)
        num_sites = diag_site_energies.size
        
        # if manifold_num == 0:
        #     alpha_aves = np.array([1 for i in range(N)])
        # elif manifold_num == 1:
        #     alpha_aves = np.array([np.mean(vib_params[i]['alphas1']) for i in range(N)])
        # elif manifold_num ==2:
        #     alpha_aves = np.array([np.mean(vib_params[i]['alphas2']) for i in range(N)])
        # w_ave = np.average(omega_gs,weights=alpha_aves)
        w_ave = np.average(omega_gs)

        num_eigs_found = 0
        quantum_number = 0
        energies_and_densities = []
        while num_eigs_found < num_eigvals:
            approx_vib_energy = w_ave * (quantum_number + N*0.5)
            approx_energies = diag_site_energies + approx_vib_energy
            number_of_nearby_vib_states = int(binom(quantum_number + N - 1,quantum_number)) + 2
            for en in approx_energies:
                energies_and_densities.append((en,number_of_nearby_vib_states))

            num_eigs_found += number_of_nearby_vib_states * num_sites
            quantum_number += 1

        ham = self.manifold_hamiltonian(manifold_num)

        eigenvalues = []
        eigenvectors = []
            
        for en, num_eigs in energies_and_densities:
            eigvals, eigvecs = eigsh(ham,k=num_eigs,which='LM',sigma=en + 1E-8)
            eigenvalues.append(eigvals)
            eigenvectors.append(eigvecs)
        print(energies_and_densities)
        print(eigenvalues)
        eigenvalues, eigenvectors = self.remove_redundant_eigenvectors(eigenvalues,eigenvectors)
        if len(eigenvalues) == 1:
            eigenvalues = eigenvalues[0][:num_eigvals]
            eigenvectors = eigenvectors[0][:,:num_eigvals]
        else:
            warnings.warn('Failure')

        for i in range(num_eigvals):
            max_index = np.argmax(np.abs(eigenvectors[:,i]))
            if eigenvectors[max_index,i] < 0:
                eigenvectors[:,i] *= -1
                
        return eigenvalues, eigenvectors
                

    def remove_redundant_eigenvectors(self,eigenvalues,eigenvectors):
        all_eigenvalues = [eigenvalues[0]]
        all_eigenvectors = [eigenvectors[0]]
        last_eigenvalues = eigenvalues[0]
        overlap_status = np.zeros(len(eigenvalues)-1,dtype=bool)
        for i in range(1,len(eigenvalues)):
            next_eigenvalues = eigenvalues[i]
            next_eigenvectors = eigenvectors[i]
            mask = np.ones(next_eigenvalues.size,dtype=bool)
            for n in range(last_eigenvalues.size):
                for m in range(next_eigenvalues.size):
                    if np.abs(last_eigenvalues[n] - next_eigenvalues[m]) < 1E-10:
                        mask[m] = False
                        overlap_status[i-1] = True
            next_eigenvalues = next_eigenvalues[mask]
            next_eigenvectors = next_eigenvectors[:,mask]
            if overlap_status[i-1]:
                all_eigenvalues[-1] = np.hstack((all_eigenvalues[-1],next_eigenvalues))
                all_eigenvectors[-1] = np.hstack((all_eigenvectors[-1],next_eigenvectors))
            else:
                all_eigenvalues.append(next_eigenvalues)
                all_eigenvectors.append(next_eigenvectors)

            last_eigenvalues = next_eigenvalues
            last_eigenvectors = next_eigenvectors

        return all_eigenvalues, all_eigenvectors
            

    def vibrationless_basis(self,manifold_num):
        elec_manifold = self.electronic_manifold(manifold_num)
        eigvals, eigvecs = eig(elec_manifold)
        sort_ind = eigvals.argsort()
        eigvals = eigvals[sort_ind]
        eigvecs = eigvecs[:,sort_ind]
        if eigvecs[0,0] < 0:
            eigvecs[:,0] *= -1
        if eigvecs[0,1] < 0:
            eigvecs[:,1] *= -1
        return eigvals, eigvecs

def SHO_basis(n,xvalues):
    coef = [0 for i in range(n)]
    coef.append(1)
    norm = 1/(np.pi**(1/4)*2**(n/2) * np.sqrt(factorial(n)))
    return  norm * np.exp(-xvalues**2/2) * hermval(xvalues,coef)

def oscillator_1d(array_1d,x):
    ans = np.zeros(x.shape,dtype=array_1d.dtype)
    for i in range(len(array_1d)):
        ans += SHO_basis(i,x) * array_1d[i]
    return ans
    
def oscillator_2d(array_2d,x1,x2):
    ans = np.zeros(x1.shape)
    l1, l2 = array_2d.shape
    for i in range(l1):
        for j in range(l2):
            ans += SHO_basis(i,x1) * SHO_basis(j,x2) * array_2d[i,j]
    return ans

def oscillator_2d_old(array_2d,x1,x2,*,rel_tol=1E-4):
    ans = np.zeros(x1.shape)
    flat_sum_ind = np.abs(array_2d.flatten()).argsort()[::-1]
    inds1, inds2 = np.unravel_index(flat_sum_ind,array_2d.shape)
    for i,j in zip(inds1, inds2):
        old_ans = ans
        ans += SHO_basis(i,x1) * SHO_basis(j,x2) * array_2d[i,j]
        if np.max(np.abs((old_ans - ans)/ans)) < rel_tol:
            break
    return ans

def plot_dimer_2vib(eigval, eigmat):
    fig = plt.figure(figsize=(12,5))
    ax1 = fig.add_axes([0.17,0.1,0.3,0.9])
    ax2 = fig.add_axes([0.55,0.1,0.3,0.9])
    cax = fig.add_axes([0.05,0.25,0.02,0.4])
    cax2 = fig.add_axes([0.9,0.25,0.02,0.4])
    ax1.set_title('Site a')
    ax2.set_title('Site b')
    fig.suptitle('E = {:.4f}'.format(eigval),fontsize=16)
    mat_a = eigmat[0,...]
    mat_b = eigmat[1,...]
    x1_1d = np.arange(-3,3,.05)
    x1, x2 = np.meshgrid(x1_1d, x1_1d)
    site_a = oscillator_2d(mat_a,x1,x2)
    site_b = oscillator_2d(mat_b,x1,x2)
    im1 = ax1.contourf(x1,x2,site_a)
    im2 = ax2.contourf(x1,x2,site_b)
    fig.colorbar(im1,cax=cax)
    fig.colorbar(im2,cax=cax2)
    ax1.set_xlabel('$q_1$',fontsize=20)
    ax1.set_ylabel('$q_2$',fontsize=20)
    ax2.set_xlabel('$q_1$',fontsize=20)
    ax2.set_ylabel('$q_2$',fontsize=20)
    ax1.set(adjustable='box-forced', aspect='equal')
    ax2.set(adjustable='box-forced', aspect='equal')
    x2a_max, x1a_max = np.unravel_index(np.argmax(np.abs(site_a)),x1.shape)
    print('Excitation on site a is centered at (q1,q2) = ({0:.2f},{1:.2f})'.format(x1_1d[x1a_max],
                                                                           x1_1d[x2a_max]))
    x2b_max, x1b_max = np.unravel_index(np.argmax(np.abs(site_b)),x1.shape)
    print('Excitation on site b is centered at (q1,q2) = ({0:.2f},{1:.2f})'.format(x1_1d[x1b_max],
                                                                           x1_1d[x2b_max]))

def plot_dimer_2vib_b(dimer_obj,eigen_index,*,fig = None):
    if fig == None:
        fig = plt.figure(figsize=(12,5))
    eigval = dimer_obj.eigenvalues[1][eigen_index]
    eigmat = dimer_obj.eigenmatrices[1][...,eigen_index]
    ax1 = fig.add_axes([0.17,0.1,0.3,0.9])
    ax2 = fig.add_axes([0.55,0.1,0.3,0.9])
    cax = fig.add_axes([0.05,0.25,0.02,0.4])
    cax2 = fig.add_axes([0.9,0.25,0.02,0.4])
    ax1.set_title('Site a')
    ax2.set_title('Site b')
    fig.suptitle('E = {:.2f}'.format(eigval),fontsize=16)
    mat_a = eigmat[0,...]
    mat_b = eigmat[1,...]
    x1_1d = np.arange(-3,3,.05)
    x1, x2 = np.meshgrid(x1_1d, x1_1d)
    site_a = oscillator_2d(mat_a,x1,x2)
    site_b = oscillator_2d(mat_b,x1,x2)
    max_val = np.max((site_a,site_b))
    min_val = np.min((site_a,site_b))
    im1 = ax1.pcolor(x1,x2,site_a,vmin=min_val,vmax=max_val)
    # ax1.contour(x1,x2,site_a)
    im2 = ax2.pcolor(x1,x2,site_b,vmin=min_val,vmax=max_val)
    # ax2.contour(x1,x2,site_b)
    d1a = dimer_obj.params['vibrations'][0]['displacements1'][0]
    d1b = dimer_obj.params['vibrations'][0]['displacements1'][1]
    wg1 = dimer_obj.params['vibrations'][0]['omega_g']
    we1a = wg1 * dimer_obj.params['vibrations'][0]['alphas1'][0]
    we1b = wg1 * dimer_obj.params['vibrations'][0]['alphas1'][1]
    wg2 = dimer_obj.params['vibrations'][1]['omega_g']
    d2a = dimer_obj.params['vibrations'][1]['displacements1'][0]
    d2b = dimer_obj.params['vibrations'][1]['displacements1'][1]
    wg2 = dimer_obj.params['vibrations'][1]['omega_g']
    we2a = wg2 * dimer_obj.params['vibrations'][1]['alphas1'][0]
    we2b = wg2 * dimer_obj.params['vibrations'][1]['alphas1'][1]
    ax1.axvline(x=d1a,c='k',ls='--')
    ax1.axhline(y=d2a,c='k',ls='--')
    ax2.axvline(x=d1b,c='k',ls='--')
    ax2.axhline(y=d2b,c='k',ls='--')
    
    fig.colorbar(im1,cax=cax)
    fig.colorbar(im2,cax=cax2)
    cax.clear()
    cax.axis('off')
    # ax1.set_xlabel('$q_1, \omega={:.2f}$'.format(we1a),fontsize=20)
    # ax1.set_ylabel('$q_2, \omega={:.2f}$'.format(we2a),fontsize=20)
    # ax2.set_xlabel('$q_1, \omega={:.2f}$'.format(we1b),fontsize=20)
    # ax2.set_ylabel('$q_2, \omega={:.2f}$'.format(we2b),fontsize=20)
    ax1.set_xlabel('$R_1$')
    ax1.set_ylabel('$R_2$')
    ax2.set_xlabel('$R_1$')
    ax2.set_ylabel('$R_2$')
    
    ax1.set(adjustable='box-forced', aspect='equal')
    ax2.set(adjustable='box-forced', aspect='equal')
    x2a_max, x1a_max = np.unravel_index(np.argmax(np.abs(site_a)),x1.shape)
    print('Excitation on site a is centered at (q1,q2) = ({0:.2f},{1:.2f})'.format(x1_1d[x1a_max],
                                                                           x1_1d[x2a_max]))
    x2b_max, x1b_max = np.unravel_index(np.argmax(np.abs(site_b)),x1.shape)
    print('Excitation on site b is centered at (q1,q2) = ({0:.2f},{1:.2f})'.format(x1_1d[x1b_max],
                                                                           x1_1d[x2b_max]))


def plot_dimer_2vib_alpha_beta(dimer_obj,eigen_index,*,fig = None):
    if fig == None:
        fig = plt.figure(figsize=(12,5))
    eigval = dimer_obj.eigenvalues[1][eigen_index]
    eigmat = dimer_obj.eigenmatrices[1][...,eigen_index]
    eigvals, eigvecs = dimer_obj.vibrationless_basis(1)
    alpha = eigvecs[:,0]
    beta = eigvecs[:,1]
    
    ax1 = fig.add_axes([0.17,0.1,0.3,0.9])
    ax2 = fig.add_axes([0.55,0.1,0.3,0.9])
    cax = fig.add_axes([0.05,0.25,0.02,0.4])
    cax2 = fig.add_axes([0.9,0.25,0.02,0.4])
    ax1.set_title(r'State $\alpha$')
    ax2.set_title(r'State $\beta$')
    fig.suptitle('E = {:.4f}'.format(eigval),fontsize=16)
    mat_alpha = eigmat[0,...]*alpha[0] + eigmat[1,...]*alpha[1]
    mat_beta = eigmat[0,...]*beta[0] + eigmat[1,...]*beta[1]
    x1_1d = np.arange(-3,3,.05)
    x1, x2 = np.meshgrid(x1_1d, x1_1d)
    site_a = oscillator_2d(mat_alpha,x1,x2)
    site_b = oscillator_2d(mat_beta,x1,x2)
    max_val = np.max((site_a,site_b))
    min_val = np.min((site_a,site_b))
    im1 = ax1.pcolor(x1,x2,site_a,vmin=min_val,vmax=max_val)
    # ax1.contour(x1,x2,site_a)
    im2 = ax2.pcolor(x1,x2,site_b,vmin=min_val,vmax=max_val)
    # ax2.contour(x1,x2,site_b)
    d1a = dimer_obj.params['vibrations'][0]['displacements1'][0]
    d1b = dimer_obj.params['vibrations'][0]['displacements1'][1]
    wg1 = dimer_obj.params['vibrations'][0]['omega_g']
    we1a = wg1 * dimer_obj.params['vibrations'][0]['alphas1'][0]
    we1b = wg1 * dimer_obj.params['vibrations'][0]['alphas1'][1]
    wg2 = dimer_obj.params['vibrations'][1]['omega_g']
    d2a = dimer_obj.params['vibrations'][1]['displacements1'][0]
    d2b = dimer_obj.params['vibrations'][1]['displacements1'][1]
    wg2 = dimer_obj.params['vibrations'][1]['omega_g']
    we2a = wg2 * dimer_obj.params['vibrations'][1]['alphas1'][0]
    we2b = wg2 * dimer_obj.params['vibrations'][1]['alphas1'][1]
    ax1.axvline(x=d1a,c='k',ls='--')
    ax1.axhline(y=d2a,c='k',ls='--')
    ax2.axvline(x=d1b,c='k',ls='--')
    ax2.axhline(y=d2b,c='k',ls='--')
    
    fig.colorbar(im1,cax=cax)
    fig.colorbar(im2,cax=cax2)
    ax1.set_xlabel('$q_1, \omega={}$'.format(we1a),fontsize=20)
    ax1.set_ylabel('$q_2, \omega={}$'.format(we2a),fontsize=20)
    ax2.set_xlabel('$q_1, \omega={}$'.format(we1b),fontsize=20)
    ax2.set_ylabel('$q_2, \omega={}$'.format(we2b),fontsize=20)
    ax1.set(adjustable='box-forced', aspect='equal')
    ax2.set(adjustable='box-forced', aspect='equal')
    x2a_max, x1a_max = np.unravel_index(np.argmax(np.abs(site_a)),x1.shape)
    # print('Excitation on site a is centered at (q1,q2) = ({0:.2f},{1:.2f})'.format(x1_1d[x1a_max],
                                                                           # x1_1d[x2a_max]))
    x2b_max, x1b_max = np.unravel_index(np.argmax(np.abs(site_b)),x1.shape)
    # print('Excitation on site b is centered at (q1,q2) = ({0:.2f},{1:.2f})'.format(x1_1d[x1b_max],
                                                                           # x1_1d[x2b_max]))
    # print('Site alpha contribution: {}'.format(np.sum(mat_alpha.flatten()**2)))
    # print('Site beta contribution: {}'.format(np.sum(mat_beta.flatten()**2)))


def plot_dimer_2vib_single_plot(eigval, eigmat):
    fig, ax = plt.subplots()
    ax.set_title('Probability density \n electronic basis traced out, E = {:.4f}'.format(eigval),fontsize=16)
    mat_a = eigmat[0,...]
    mat_b = eigmat[1,...]
    x1_1d = np.arange(-3,3,.05)
    x1, x2 = np.meshgrid(x1_1d, x1_1d)
    site_a = oscillator_2d(mat_a,x1,x2)
    site_b = oscillator_2d(mat_b,x1,x2)
    total = site_a**2 + site_b**2
    im = ax.contourf(x1,x2,total)
    fig.colorbar(im)
    ax.set_xlabel('$q_1$',fontsize=20)
    ax.set_ylabel('$q_2$',fontsize=20)
    ax.set(adjustable='box-forced', aspect='equal')
    x2a_max, x1a_max = np.unravel_index(np.argmax(total),x1.shape)
    print('Excitation is centered at (q1,q2) = ({0:.2f},{1:.2f})'.format(x1_1d[x1a_max],
                                                                           x1_1d[x2a_max]))
    print('Site a contribution: {}'.format(np.sum(mat_a.flatten()**2)))
    print('Site b contribution: {}'.format(np.sum(mat_b.flatten()**2)))
    
    
def plot_dimer_2vib_g_or_f(eigval, eigmat):
    fig, ax1 = plt.subplots()
    ax1.set_title('E = {:.4f}'.format(eigval),fontsize=16)
    mat_a = eigmat[0,...]
    x1_1d = np.arange(-3,3,.05)
    x1, x2 = np.meshgrid(x1_1d, x1_1d)
    site_a = oscillator_2d(mat_a,x1,x2)
    im1 = ax1.contourf(x1,x2,site_a)
    fig.colorbar(im1)
    ax1.set_xlabel('$q_1$',fontsize=20)
    ax1.set_ylabel('$q_2$',fontsize=20)
    ax1.set(adjustable='box-forced', aspect='equal')
    x2a_max, x1a_max = np.unravel_index(np.argmax(np.abs(site_a)),x1.shape)
    print('Excitation on site a is centered at (q1,q2) = ({0:.2f},{1:.2f})'.format(x1_1d[x1a_max],
                                                                           x1_1d[x2a_max]))

def plot_dimer_singly_excited_states(obj,*,plot_range=[-0.1,4.1],fig=None):
    if fig == None:
        fig = plt.figure()
    ax = fig.add_axes([0.1,0.1,0.8,0.8])
    for ev in obj.eigenvalues[1]:
        ax.axvline(x=ev)
    ax.set_xlim(plot_range)
    ax.set_xlabel('Energy ($\hbar\omega$)')
    d = (obj.params['vibrations'][0]['displacements2'][0] +
         obj.params['vibrations'][1]['displacements2'][0] )/2
    ax.set_title('Eigenenergies for $J =$ {:.2f}, $d =$ {:.2f}'.format(obj.couplings[1][0],d))
        
